#ifndef __3DPATH_H__
#define __3DPATH_H__

#include "f3Dutils.h"
#include <vector>

struct pos_t {
	fVector eye;
	fVector ref;
	fVector up;
	Uint32 time;
};

class Path
{
 private:
	vector<pos_t> path;
	unsigned int lastpos;
	
 public:
	Path() {
		lastpos=0;
	}
	void push(pos_t pos) {
		path.push_back(pos);
	}
	void push(Uint32 time,
			  float xeye, float yeye, float zeye,
			  float xref, float yref, float zref,
			  float xup, float yup, float zup) {
		pos_t pos;
		pos.time = time;
		pos.eye[0] = xeye; pos.eye[1] = yeye; pos.eye[2] = zeye;
		pos.ref[0] = xref; pos.ref[1] = yref; pos.ref[2] = zref;
		pos.up[0] = xup; pos.up[1] = yup; pos.up[2] = zup;
		path.push_back(pos);
	}
	bool getpos(Uint32 time, pos_t &pos){
		// no pos
		if(!path.size()) return false;
		// after last pos !
		if(time > path[path.size()-1].time) { return false; }
		// before first pos !
		if(time < path[lastpos].time) { return false; }
		// if current element ok :)
		if(time == path[lastpos].time) {
			pos = path[lastpos];
			return true;
		}
		// ended
		if(lastpos == path.size()-1) { return false; }
		// if next element ok
		if(time == path[lastpos+1].time) {
			lastpos++;
			pos = path[lastpos];
			return true;
		}
		// next pos time < last ?!?!?!
		if(path[lastpos].time > path[lastpos+1].time) { return false; }
		// update lastpos
		if(time > path[lastpos+1].time) lastpos++;
		if(lastpos == path.size()-1) return false;
		// interpolate
		float delta = (float)(time - path[lastpos].time)/(float)(path[lastpos+1].time-path[lastpos].time);
		pos.eye[0] = path[lastpos].eye[0] + delta * (path[lastpos+1].eye[0]-path[lastpos].eye[0]);
		pos.eye[1] = path[lastpos].eye[1] + delta * (path[lastpos+1].eye[1]-path[lastpos].eye[1]);
		pos.eye[2] = path[lastpos].eye[2] + delta * (path[lastpos+1].eye[2]-path[lastpos].eye[2]);
		pos.ref[0] = path[lastpos].ref[0] + delta * (path[lastpos+1].ref[0]-path[lastpos].ref[0]);
		pos.ref[1] = path[lastpos].ref[1] + delta * (path[lastpos+1].ref[1]-path[lastpos].ref[1]);
		pos.ref[2] = path[lastpos].ref[2] + delta * (path[lastpos+1].ref[2]-path[lastpos].ref[2]);
		pos.up[0] = path[lastpos].up[0] + delta * (path[lastpos+1].up[0]-path[lastpos].up[0]);
		pos.up[1] = path[lastpos].up[1] + delta * (path[lastpos+1].up[1]-path[lastpos].up[1]);
		pos.up[2] = path[lastpos].up[2] + delta * (path[lastpos+1].up[2]-path[lastpos].up[2]);
		return true;
	}
};

#endif //__3DPATH_H__
