#ifndef __RENDER_H__
#define __RENDER_H__

#include <stdlib.h>
#include <map>
#include <string>

#include "SDL.h"
#include "f3Dutils.h"

struct Poly {
	int n;
	int z;
	fVector *vertex; // vertex
	iVector *pvertex; // projected vertex
	iVector *texel;
	int texture;
	int flags;
	bool draw;
};

#define VSIZE(NB) (3*4*NB)
#define FSIZE(NB) (6*4*NB)

class Render
{
private:
	Uint32 *screen;
	map<int,SDL_Surface*> textures;
	map<string,int> tex_byname;

	int mem_alloc; // bytes allocated
	int mem_used; // bytes used
	bool mem_overflow; // overflow (mem_used > mem_alloc)
	void *mem;

	int poly_max;
	int poly_count;
	Poly *polys;
	
	// context
	int current_texture;
	fMatrix cam;
	fVector eye;

	int model_max;
	int model_count;
	void *modelstack;
	
 public:
	Render();
	~Render();

	// cherche fichier dans PATH_2D
	int bind(char *filename);
	int bind(int texture);
	SDL_Surface *getTexture(int texture);

	void camera(fMatrix cam);
	void set_screen(Uint32 *screen);
	void LookAt(fVector &eye, fVector &ref, fVector &up);
	void LookAt(fVector &eye, fVector &ref, 
		    float angle);

	void *allocVector(int n);
	fVector *allocfVector(int n, fVector *fv);
	Poly *allocPoly();
	// n = nb de vertex // flags : CULL_FACE
	void push(int n, fVector *vertex, fVector *texel, int texture, int flags);
	void project();
	void zsort();
	void go();
	void flush_polys();

	fMatrix modelview;
	void push_matrix();
	void pop_matrix();

	void translate(float x, float y, float z) {
		fVector tv;
		f_make_vector(tv,x,y,z);
		translate(tv);
	}
	void translate(fVector &tv) {
		fMatrix m,tm;
		f_make_trans_matrix(tm,tv);
		f_MatrixCopy(modelview,m);
		f_ConcatMatrices(m,tm,modelview);
	}
	void scale(float x, float y, float z) {
		fVector sv;
		f_make_vector(sv,x,y,z);
		scale(sv);
	}
	void scale(fVector &sv) {
		fMatrix m,sm;
		f_make_scale_matrix(sm,sv);
		f_MatrixCopy(modelview,m);
		f_ConcatMatrices(m,sm,modelview);
	}
	void rotate(float xa, float ya, float za) {
		fMatrix rm,m;
		f_make_identity_matrix(rm);
		f_append_xrot_matrix(rm,xa);
		f_append_yrot_matrix(rm,ya);
		f_append_zrot_matrix(rm,za);
		f_MatrixCopy(modelview,m);
		f_ConcatMatrices(m,rm,modelview);
	}
};

#endif /* __RENDER_H__ */
