#include "Object.h"
#include "effets.h"

Object::Object(char *filename, Render *rd, float scalefactor)
{
	// Keep the render context
	this->rd = rd;
	this->scalefactor = scalefactor;
	
	// load the file
	file = lib3ds_file_load(filename);
	if (!file) {
		cout << "[** ERROR **] loading 3ds file: " << filename << endl;
		return;
	}

	// load textures
	int npolys = 0;
	for (Lib3dsMesh *mesh=file->meshes; mesh; mesh=mesh->next) {
		npolys+=mesh->faces;
		for (unsigned int p=0; p<mesh->faces; ++p) {
			Lib3dsFace *f=&mesh->faceL[p];
			Lib3dsMaterial *mat = lib3ds_file_material_by_name(file, f->material);
			if( mat && mat->texture1_map.name )
				f->user.i = rd->bind(mat->texture1_map.name);
			else f->user.i = -1;
		}
	}
	cout << "3DS ["<<filename<<"] nb polygones: "<< npolys << endl;
}

Object::~Object()
{
	if(file) lib3ds_file_free(file);
}

void Object::render()
{
	if(!file) return;

	fVector toto[3];
	fVector ntri[3];
	for (Lib3dsMesh *mesh=file->meshes; mesh; mesh=mesh->next) {
		if(!mesh->texelL) {
			//cout << "Missing texture mapping !!!" << endl;
			continue;
		}
		for (unsigned int p=0; p<mesh->faces; ++p) {
			Lib3dsFace *f=&mesh->faceL[p];
			for(int i=0; i<3; i++ ) {
				f_VectCopy(mesh->pointL[f->points[i]].pos,ntri[i]);
				toto[i][0] = mesh->texelL[f->points[i]][0];
				toto[i][1] = mesh->texelL[f->points[i]][1];
			}
			if(f->user.i!=-1) {
				if(scalefactor!=1.0) {
					rd->push_matrix();
					rd->scale(scalefactor,scalefactor,scalefactor);
					rd->push(3,ntri,toto,f->user.i,0);
					rd->pop_matrix();
				}
				else rd->push(3,ntri,toto,f->user.i,0);
			}
		}
	}
}
