#include "effets.h"
#include "Render.h"
#include "quat.h"

static int mysort(const void *poly1, const void *poly2)
{
	return ((Poly*)poly1)->z-((Poly*)poly2)->z;
}

void Render::set_screen(Uint32 *screen)
{
	this->screen = screen;
}

Render::Render()
{
	screen = NULL;
	mem_alloc = 1000000;
	mem_used = 0;
	mem_overflow = false;
	mem = malloc(mem_alloc);

	poly_max = 10000;
	poly_count = 0;
	polys = (Poly*)malloc(poly_max*sizeof(Poly));

	model_max = 100;
	model_count = 0;
	modelstack = (void*)malloc(sizeof(fMatrix)*model_max);

	current_texture = -1;

	f_make_identity_matrix(modelview);
	f_make_identity_matrix(cam);
}

Render::~Render()
{
	free(polys);
	free(mem);
}

void Render::flush_polys()
{
	if(mem_overflow) {
		cout << "Render::flush_polys() mem_overflow alloc:"<<mem_alloc<<" used:"<<mem_used<<endl;
		mem_alloc = mem_used;
		mem = realloc(mem,mem_alloc);
		mem_overflow = false;
	}
	poly_count = 0;
	mem_used = 0;
	model_count = 0;
}

static Uint32 ConvertPixel(const Uint32 src,  const Uint32 srcformat, 
						   const Uint32 srcshift, const Uint32 SHIFT)
{
	return ( (src&srcformat) >> srcshift ) << SHIFT;
}

// cherche fichier dans PATH_2D
int Render::bind(char *filename)
{
	if(!filename || !strlen(filename)) return -1;
	
	map<string,int>::iterator t = tex_byname.find(string(filename));
	if(t!=tex_byname.end()) return t->second;	
	
	string tmpfname = PATH_2D;
	tmpfname+=filename;
	cout << "** Loading texture : "<< tmpfname.c_str() << endl;
	SDL_Surface *load = IMG_Load(tmpfname.c_str());

	if( !load ) { printf("La texture est pas la !!!!\n"); return -1; }

	SDL_Surface *mytexture = \
		SDL_CreateRGBSurface ( SDL_SWSURFACE, load->w, load->h, 32,
							   0xFF0000, 0xFF00, 0xFF, 0xFF000000);

	if( load->format->BytesPerPixel != 4) {
		printf("  '-> Alpha missing\n");
		SDL_BlitSurface(load,NULL, mytexture, NULL);
		
	} else {
		printf("  '-> Alpha present :-)\n");
	
		Uint32 *src = (Uint32 *)load->pixels;
		Uint32 *dst = (Uint32 *)mytexture->pixels;
	
		for (int i=0; i<load->w*load->h; i++) {
			SDL_PixelFormat &sf = *(load->format);
			int point = 
				ConvertPixel(src[i], sf.Rmask, sf.Rshift, RSHFT) + 
				ConvertPixel(src[i], sf.Gmask, sf.Gshift, GSHFT) +
				ConvertPixel(src[i], sf.Bmask, sf.Bshift, BSHFT) +
				ConvertPixel(src[i], sf.Amask, sf.Ashift, ASHFT);
			dst[i] = point;
		}
	}
	SDL_FreeSurface(load);
	
	int i=1;
	do if( textures.find(i) == textures.end() ) {
		textures[i] = mytexture;
		tex_byname[string(filename)] = i;
		//cout << "** registered " << i << endl;
		return i;
	} while (i++);
	return -1;
}

int Render::bind(int texture)
{
	if( textures.find(texture) != textures.end() ) {
		current_texture = texture;
		return texture;
	}
	else return -1;
}

SDL_Surface *Render::getTexture(int texture)
{
	if( textures.find(texture) != textures.end() ) {
		return textures[texture];
	}
	else return NULL;
}

void Render::camera(fMatrix cam)
{
	f_MatrixCopy(cam, this->cam);
}

void *Render::allocVector(int n)
{
	void *ptr;
	if(mem_alloc -mem_used < VSIZE(n)) {
		mem_overflow = true;
		ptr = NULL;
	} else {
		ptr = (unsigned char*)mem+mem_used;
	}
	mem_used+=VSIZE(n);
	return ptr;
}

fVector *Render::allocfVector(int n, fVector *fv)
{
	void *ptr = allocVector(n);
	if(ptr) memcpy(ptr,fv,VSIZE(n));
	return (fVector*)ptr;
}

Poly *Render::allocPoly()
{
	if(poly_max-poly_count < 2) {
		poly_max+=500;
		polys = (Poly*)realloc(polys,poly_max*sizeof(Poly));
	}
	Poly *p = polys+poly_count;
	poly_count++;
	return p;
}

// n = nb de vertex // flags : CULL_FACE
void Render::push(int n, fVector *vertex, fVector *texel, int texture, int flags)
{
	Poly *p = allocPoly();
	p->n = n;
	p->vertex = allocfVector(n,vertex);
	if(texel) p->texel = (iVector*)allocfVector(n,(fVector*)texel);
	p->texture = texture;
	p->flags = flags;
	if(n>1 && bind(texture)==-1) {
		p->draw = false;
		return;
	} else p->draw = true;

	if(p->vertex) {
		if(p->texel && p->texture) {
			iVector *iv = p->texel;
			for(int i=0; i<n; i++ ){
				iv[i][0] = (int)(texel[i][0] * 65536.0f * textures[texture]->w);
				iv[i][1] = ((textures[texture]->h)<<16) - (int)(texel[i][1] * 65536.0f * textures[texture]->h);
			}
		} else if(p->n > 1)	p->draw = false;
		while(n--) f_XformVector(modelview,p->vertex[n],p->vertex[n]);
	}
}

void Render::project()
{
	// regarde si on a assez de memoire
	if(mem_overflow || mem_used*3 > mem_alloc*2) {
		//cout << "Render::project() ==> mem_overflow" << endl;
		mem_used = (mem_used*3)/2;
		mem_overflow = true;
		return;
	}
	const float CLIPZ = 0.001;
	int n = poly_count;
	while(n--) {
		for(int i=0; i<polys[n].n; i++) {
			f_XformVector(cam, polys[n].vertex[i], polys[n].vertex[i]);
		}
		fVector *ntri = polys[n].vertex;
		iVector *intri = polys[n].pvertex = (iVector*)allocVector(polys[n].n);
		int i,depth = 0;
		if(!intri) { mem_overflow=true; goto z_skip; }
		for (i=0; i<polys[n].n; i++) {
			if( ntri[i][2] <= CLIPZ ) goto z_skip;
			float z = 400.0 / (ntri[i][2] + 10.0);
			ntri[i][0] *= z;
			ntri[i][1] *= z;
			
			for (int j=0; j<3; j++) {
				intri[i][j] = (int) (ntri[i][j]*65536.0);	
			}
			depth+=intri[i][2];
		}
		polys[n].z=depth/i;
		continue;
	z_skip:
		polys[n].draw = false;
		continue;
	}
}

void Render::zsort()
{
	if(mem_overflow) {
		//cout << "Render::zsort() ==> mem_overflow" << endl;
		return;
	}
	qsort(polys, poly_count, sizeof(Poly), mysort);
}

void Render::go()
{
	if(mem_overflow) return;
	if(!screen) return;
	int n = poly_count;
	while(n--) {
		if(polys[n].draw){
		  
			if(polys[n].n > 1) {
				render_tri( screen, polys[n].pvertex, polys[n].texel,
							(Uint32*)(textures[polys[n].texture]->pixels),
							textures[polys[n].texture]->w,
							textures[polys[n].texture]->h,
							polys[n].flags, polys[n].n );
			} else if(parts::getInstance()) {
				asm_emms();
				parts::getInstance()->putPart2( screen, (int*)polys[n].pvertex );
			}
		}
	}
	asm_emms();
}

void Render::push_matrix()
{
	if( model_count > model_max ) {
		model_max += 100;
		modelstack = (void*)realloc(modelstack,model_max*sizeof(fMatrix));
	}
	memcpy(((unsigned char*)modelstack) + sizeof(fMatrix)*model_count,modelview,sizeof(fMatrix));
	model_count++;
}

void Render::pop_matrix()
{
	if(!model_count) return;
	model_count--;
	memcpy(modelview,((unsigned char*)modelstack) + sizeof(fMatrix)*model_count,sizeof(fMatrix));
}

void Render::LookAt(fVector &eye, fVector &ref, fVector &up)
{
  fVector Z,X,Y;
  
  f_VectSub(ref,eye,Z);
  f_NormalizeVect(Z);
  f_CrossProduct(up,Z,X);
  f_NormalizeVect(X);
  f_CrossProduct(Z,X,Y);
  f_make_identity_matrix(cam);
  f_VectCopy(X,cam);
  f_VectCopy(Y,cam+4);
  f_VectCopy(Z,cam+8);
  
  f_VectCopy(eye, this->eye);
  f_make_identity_matrix(modelview);
  f_trans_matrix(modelview,-eye[0],-eye[1],-eye[2]);
}

// marche pas
void Render::LookAt(fVector &eye, fVector &ref, 
		    float angle)
{
  Quat quat;
  fVector v2;

  f_VectSub(ref,eye,v2);  
  f_NormalizeVect(v2);

  /*  float r = 1.0 /
    sqrt(ref[0]*ref[0] + ref[1]*ref[1] + ref[2]*ref[2]);
    v2[0] = ref[0]*r; v2[1] = ref[1]*r; v2[2] = ref[2]*r;*/
  
  quat.CreateFromAxis(v2[0], v2[1], v2[2], angle);
  quat.ToMatrix(cam);
  f_transpose_matrix(cam);
  // f_make_identity_matrix(cam);

  f_VectCopy(eye, this->eye);
  f_make_identity_matrix(modelview);
  f_trans_matrix(modelview, -eye[0], -eye[1], -eye[2]);
}
