/* Etoile */
/* Une toile qui brule en faisant un fire */

#include <math.h>
#include <assert.h>
#include "effets.h"
#include "f3Dutils.h"
#include "sable1.h"
#include "Render.h"
#include "etoile.h"

#include "quat.h"
#include "plan.h"

Uint32 Etoile::cmerge(const int src, const int dst, const int ALPHA)
{
        const int ret = ( src*(ALPHA) + dst*(255-ALPHA) ) >> 8;
        //assert (ret >= 0 && ret <= 255);        // to be removed
        return ret;
}


void Etoile::LowDrawAlpha(Uint32 *screen, Uint32 *soff,
		  const int ycount, const int xcount)
{
  for (int y=0; y<ycount; y++) {
    for (int x=0; x<xcount; x++) {
      const Uint32 sc = *soff;    // image (source)
      const Uint32 dc = *screen;                // screen (dest)
      Uint32 Rsc = (sc&RMASK) >> RSHFT; Uint32 Gsc = (sc&GMASK) >> GSHFT; 
      Uint32 Bsc = (sc&BMASK) >> BSHFT; 
      const Uint32 ALPHA = (sc&AMASK) >> ASHFT;
      
      Uint32 Rdc = (dc&RMASK) >> RSHFT; Uint32 Gdc = (dc&GMASK) >> GSHFT; 
      Uint32 Bdc = (dc&BMASK) >> BSHFT; 
      
      Uint32 Rfc = cmerge(Rsc, Rdc, ALPHA)<<RSHFT;
      Uint32 Gfc = cmerge(Gsc, Gdc, ALPHA)<<GSHFT;
      Uint32 Bfc = cmerge(Bsc, Bdc, ALPHA)<<BSHFT;
      
      const Uint32 ColorFinal = Rfc + Gfc + Bfc;
      
      *screen = ColorFinal;
      screen++; soff++;
    }
  }
}

Etoile::Etoile(Plan *plan_, vector<Vect> pos_, 
	       vector<Vect> rot_, Sable &sable1)
{

  for (int i=0; i<NBS*5; i++) {
    bruit[i] = random()%4000;
  }

  for (int i=0; i<NBS*5; i++) {
    tsable[i] = new Sable(sable1);
  }

  const float TOUR = 2.0*M_PI;

  pos = pos_;
  rot = rot_;

  x_star = new float[NBSTAR];
  y_star = new float[NBSTAR];
  vstar = new fVector[NBSTAR];

  buff = new Uint32[SCREENWIDTH*SCREENHEIGHT];

  // ligne: src -> dest
  lsrc = new char[NBSTAR+1];

  int j = 0;
  for (float i=0; j<NBSTAR; i += TOUR/NBSTAR, j++) {
    const float x = R * cos(i) ;
    const float y = R * sin(i) ;
    vstar[j][0] = x;
    vstar[j][1] = y;
    vstar[j][2] = 0;
  }
  plan = plan_;
}

Etoile::~Etoile()
{
  delete[] x_star;
  delete[] y_star;
  delete[] lsrc;
  delete[] buff;
}

void Etoile::go( Uint32* buffer, Uint32 time, bool debug_cam ) 
{

  fMatrix M;

  fVector rotn;
  fVector posn;

  const unsigned int MAX = rot.size();

  assert (MAX == pos.size());
  if (pos.size() == 0) return;

  const int TWOPART = 800;

  int p = (time/TWOPART) % MAX;
  int p2 = (p + 1)%MAX;

  Quat quat;
  if (debug_cam) {
    plan->getMouse(posn, rotn);
    quat.CreateFromEuler(rotn[0], rotn[1], rotn[2]);
  }
  else {
    Quat quat1;
    quat1.CreateFromEuler(rot[p].x, rot[p].y, rot[p].z);
    quat.CreateFromEuler(rot[p2].x, rot[p2].y, rot[p2].z);  
    const float T = (float) (time%TWOPART)/ (float) TWOPART;
    quat.Slerp(quat1, quat, T);

    for (int i=0; i<3; i++) {
      posn[i] = pos[p].v[i]*(1-T) + pos[p2].v[i]*T;
    }
  }

  quat.ToMatrix(M);

  for (int i=0; i<NBSTAR; i++) {
    fVector dv;
    f_XformVector(M, vstar[i], dv);
    float z = 1024.0/(dv[2]+posn[2]+1024);
    if (z>0.0) {
      float x = (dv[0]+posn[0]) * z;
      float y = (dv[1]+posn[1]) * z;
      x += SCREENWIDTH/2;
      y += SCREENHEIGHT/2;
      x_star[i] = x;
      y_star[i] = y;
    }
    else {
      x_star[i] = SCREENWIDTH+1;
      y_star[i] = SCREENHEIGHT+1;
    }
  }

  int sableOffset = 0;
  for (int i=0; i<NBSTAR; i++) {
    int x = (int) x_star[i];
    int y = (int) y_star[i];

    char ldst = (i+2)%NBSTAR;
    int x2 = (int) x_star[ldst];
    int y2 = (int) y_star[ldst];
    float a = time*0.001;
    int toto = 150 + (int) (100.0*cos(a));
    int toto2 = 128 + (int) (128.0*cos(a+M_PI/2));


    float dx = (x2 - x)/NBS;
    float dy = (y2 - y)/NBS;
    float fx = x-SCREENWIDTH/2; 
    float fy = y-SCREENHEIGHT/2;

    for (int a=0; a<NBS; a++) {
      sableOffset = (sableOffset+1) % (NBS*5);
      Sable *sa = tsable[sableOffset];
      int cx = (int) fx;
      int cy = (int) fy;
      sa->setcx(cx);
      sa->setcy(cy);
      sa->go(time +bruit[(sableOffset+time) % (NBS*5) ], buff);

      fy += dy;
      fx += dx;
    }

    line (buff, x, y, x2, y2, (toto<<RSHFT)+ (toto2>>2) );

    if (x>=0 && x<SCREENWIDTH && y>=0 && y<SCREENHEIGHT) {
      int offset = x + y*SCREENWIDTH;
      buff[offset] = 0xFFFFFF;
    }
  }
  blur(buff,buffer);
}

inline
bool Etoile::xin(int x)
{
  return (x>=0 && x<SCREENWIDTH);
}
inline
bool Etoile::yin(int y)
{
  return (y>=0 && y<SCREENHEIGHT);
}

void Etoile::blur (Uint32 *src, Uint32 *dst)
{


  int offdest = 0;
  for (int y=0; y<SCREENHEIGHT; y++) {
    for (int x=0; x<SCREENWIDTH; x++) {
      const int A = 1;
      const int B = 2;
      const int C = 4;
      int x1 = x - A;
      int x2 = x + A;
      int y1 = y + B;
      int y2 = y + C;
      if ( xin(x1) && xin(x2) && yin(y1) && yin(y2) ) {
	Uint32 bc = 
	  (src[offdest+A]&BMASK) +
	  (src[offdest-A]&BMASK) +
	  (src[offdest+B*SCREENWIDTH]&BMASK) +
	  (src[offdest+C*SCREENWIDTH]&BMASK);
	Uint32 gc = 
	  (src[offdest-A]&GMASK) +
	  (src[offdest+A]&GMASK) +
	  (src[offdest+B*SCREENWIDTH]&GMASK) +
	  (src[offdest+C*SCREENWIDTH]&GMASK);
	Uint32 rc = 
	  (src[offdest-A]&RMASK) +
	  (src[offdest+A]&RMASK) +
	  (src[offdest+B*SCREENWIDTH]&RMASK) +
	  (src[offdest+C*SCREENWIDTH]&RMASK);

	const int ATT = 15;

	Uint32 c = 
	  (((bc*ATT)>>6)&BMASK) +
	  (((gc*ATT)>>6)&GMASK) +
	    (((rc*ATT)>>6)&RMASK);

	dst[offdest] = c;
      }
      else {
	dst[offdest] = 0;
      }
      offdest++;
    }
  }
  for (int i=0; i<SCREENWIDTH*SCREENHEIGHT; i++) {
    src[i] = dst[i];
  }
}

void Etoile::line(Uint32 *screen, int xd, int yd, 
		  int xf, int yf, Uint32 col)
{
  int x,y,delta, index, off;
  int xx, yy;

  if (abs(xf-xd) > abs(yf-yd))
  {
  if (xd > xf) {swap(xd, xf);
               swap(yd, yf);}
  
  if (yf-yd == 0)
    { off=yd*640+xd;
    for (x=xd; x < xf; x++)
    {
      if ((x > 0) && (x < 640) && (yd > 0) && (yd < 480))
          *(screen+off)=col+bruit[x%NBS*5];
      off++;
    }
     
    }
  else {
  delta=(yf-yd)*256/(xf-xd);
  index=yd*256;
  for (x=xd; x<xf; x++)
    {   yy=(index / 256);
        off=yy*640+x;
                if ((x > 0) && (x < 640) && (yy > 0) && (yy < 480))
               *(screen+off)=col+bruit[x%NBS*5];
       index+=delta;       
    }
  }  
  }  
     
     
  if (abs(xf-xd) <= abs(yf-yd))
  {
  if (yd > yf) {swap(xd, xf);
               swap(yd, yf);}

  if (xf-xd == 0)
  { off=yd*640+xd;
      for (y=yd; y< yf; y++)
      {
        if ((xd > 0) && (xd < 640) && (y > 0) && (y < 480))
            *(screen+off)=col+bruit[x%NBS*5];
        off+=640;
      }
  }
  else {
  delta=(xf-xd)*256/(yf-yd);
  index=xd*256;
  xx=xd;
  for (y=yd; y<yf; y++)
    {     xx=index/256; 
          off= xx+y*640; 
                if ((xx > 0) && (xx < 640) && (y > 0) && (y < 480))
            *(screen+off)=col+bruit[x%NBS*5];
       index+=delta;         
    }
  }  
  }  
}    
