/*
 *  MATRIX
 */
#define FMATRIX_C
#include "effets.h"

void f_printmatrix( fMatrix m )
{
	printf("m [ %f, %f, %f, %f ]\n", m[0], m[1], m[2],  m[3]);
	printf("m [ %f, %f, %f, %f ]\n", m[4], m[5], m[6],  m[7]);
	printf("m [ %f, %f, %f, %f ]\n\n", m[8], m[9], m[10], m[11]);
}

float f_identity_matrix[12] =
{
    1.0,   0.0,   0.0,   0.0,
    0.0,   1.0,   0.0,   0.0,
    0.0,   0.0,   1.0,   0.0,
};

void f_make_identity_matrix(fMatrix m)
{
    memcpy(m, f_identity_matrix, 48);
}

void f_make_scale_matrix(fMatrix m, fVector sv)
{
    *m++ = *sv++;  *m++ =   0.0;  *m++ = 0.0;  *m++ = 0.0;
    *m++ =   0.0;  *m++ = *sv++;  *m++ = 0.0;  *m++ = 0.0;
    *m++ =   0.0;  *m++ =   0.0;  *m++ = *sv;  *m   = 0.0;
}

void f_make_trans_matrix(fMatrix m, fVector tv)
{
    *m++ = 1.0;  *m++ = 0.0;  *m++ = 0.0;  *m++ = *tv++;
    *m++ = 0.0;  *m++ = 1.0;  *m++ = 0.0;  *m++ = *tv++;
    *m++ = 0.0;  *m++ = 0.0;  *m++ = 1.0;  *m   = *tv;
}

// gnre une matrice de rotation directement avec les 3 angles
// dtruit la partie transition
void f_make_xyzrot_matrix(fMatrix m, fIangle ax, fIangle ay, fIangle az)
{

  const float A = cos(ax);
  const float B = sin(ax);
  const float C = cos(ay);
  const float D = sin(ay);
  const float E = cos(az);
  const float F = sin(az);
  const	float AD= A*D;
  const	float BD= B*D;

  m[0]=  C*E;         m[1]= -C*F;         m[2]= -D; 
  m[4]= -BD*E + A*F;  m[5]=  BD*F + A*E;  m[6]= -B*C;
  m[8]=  AD*E + B*F;  m[9]= -AD*F + B*E;  m[10]= A*C; 

  m[3] = m[7] = m[11] = 0;
}

void f_make_xrot_matrix(fMatrix m, fIangle theta)
{
    float trig_sin, trig_cos;

    //f_CosSin(theta, &trig_cos, &trig_sin);
    trig_sin = sin( theta );
    trig_cos = cos( theta );

    *m++ = 1.0;  *m++ =       0.0;  *m++ =       0.0;  *m++ = 0.0;
    *m++ = 0.0;  *m++ =  trig_cos;  *m++ =  trig_sin;  *m++ = 0.0;
    *m++ = 0.0;  *m++ = -trig_sin;  *m++ = -trig_cos;  *m   = 0.0;
}

void f_make_yrot_matrix(fMatrix m, fIangle theta)
{
    float trig_sin, trig_cos;

    //f_CosSin(theta, &trig_cos, &trig_sin);
    trig_sin = sin( theta );
    trig_cos = cos( theta );

    *m++ =  trig_cos;  *m++ = 0.0;  *m++ = -trig_sin;  *m++ = 0.0;
    *m++ =       0.0;  *m++ = 1.0;  *m++ =       0.0;  *m++ = 0.0;
    *m++ = -trig_sin;  *m++ = 0.0;  *m++ =  trig_cos;  *m   = 0.0;
}

void f_make_zrot_matrix(fMatrix m, fIangle theta)
{
    float trig_sin, trig_cos;

    //f_CosSin(theta, &trig_cos, &trig_sin);
    trig_sin = sin( theta );
    trig_cos = cos( theta );

    *m++ =  trig_cos;  *m++ =  trig_sin;  *m++ = 0.0;  *m++ = 0.0;
    *m++ = -trig_sin;  *m++ =  trig_cos;  *m++ = 0.0;  *m++ = 0.0;
    *m++ =       0.0;  *m++ =       0.0;  *m++ = 1.0;  *m   = 0.0;
}

void f_MatrixCopy(fMatrix sm, fMatrix dm)
{
    memcpy(dm, sm, 48);
}

void f_ConcatMatrices(fMatrix m1, fMatrix m2, fMatrix dm)
{
    int i, temp;

    temp = 0;
    i = 3;
    while (i--)
    {
        *dm++ = (m1[temp  ] * m2[0]) +
                (m1[temp+1] * m2[4]) +
                (m1[temp+2] * m2[8]);

        *dm++ = (m1[temp  ] * m2[1]) +
                (m1[temp+1] * m2[5]) +
                (m1[temp+2] * m2[9]);

        *dm++ = (m1[temp  ] * m2[2]) +
                (m1[temp+1] * m2[6]) +
                (m1[temp+2] * m2[10]);

        *dm++ = (m1[temp  ] * m2[3]) +
                (m1[temp+1] * m2[7]) +
                (m1[temp+2] * m2[11]) + m1[temp+3];

        temp += 4;
    }
}

void f_trans_matrix(fMatrix m, fVector dv)
{
    m[ 3] += *dv++;
    m[ 7] += *dv++;
    m[11] += *dv;
}

void f_trans_matrix(fMatrix m, float x, float y, float z )
{
    m[ 3] += x;
    m[ 7] += y;
    m[11] += z;
}

void f_transpose_matrix(fMatrix &m)
{
	fMatrix n;

	for (int i=0; i<12; i++) {
		n[i] = m[i];
	}

	m[0] = n[0]; m[1] = n[4]; m[2] = n[8];
	m[4] = n[1]; m[5] = n[5]; m[6] = n[9];
	m[8] = n[2]; m[9] = n[6]; m[10]= n[10];
}

void f_append_xrot_matrix(fMatrix m, fIangle theta)
{
    float trig_sin, trig_cos;
    float temp0, temp1, temp2;

    //f_CosSin(theta, &trig_cos, &trig_sin);
    trig_sin = sin( theta );
    trig_cos = cos( theta );

    temp0 = (trig_cos * m[4]) + (-trig_sin * m[8]);
    temp1 = (trig_cos * m[5]) + (-trig_sin * m[9]);
    temp2 = (trig_cos * m[6]) + (-trig_sin * m[10]);

    m[8]  = (trig_sin * m[4]) + (trig_cos * m[8]);
    m[9]  = (trig_sin * m[5]) + (trig_cos * m[9]);
    m[10] = (trig_sin * m[6]) + (trig_cos * m[10]);

    m[4]  = temp0;
    m[5]  = temp1;
    m[6]  = temp2;
}

void f_append_yrot_matrix(fMatrix m, fIangle theta)
{
    float trig_cos, trig_sin;
    float temp0, temp1, temp2;

    //f_CosSin(theta, &trig_cos, &trig_sin);
    trig_sin = sin( theta );
    trig_cos = cos( theta );

    temp0 = (trig_cos * m[0]) + (trig_sin * m[8]);
    temp1 = (trig_cos * m[1]) + (trig_sin * m[9]);
    temp2 = (trig_cos * m[2]) + (trig_sin * m[10]);

    m[8]  = (-trig_sin * m[0]) + (trig_cos * m[8]);
    m[9]  = (-trig_sin * m[1]) + (trig_cos * m[9]);
    m[10] = (-trig_sin * m[2]) + (trig_cos * m[10]);

    m[0]  = temp0;
    m[1]  = temp1;
    m[2]  = temp2;
}

void f_append_zrot_matrix(fMatrix m, fIangle theta)
{
    float trig_cos, trig_sin;
    float temp0, temp1, temp2;

    //f_CosSin(theta, &trig_cos, &trig_sin);
    trig_sin = sin( theta );
    trig_cos = cos( theta );

    temp0 = (trig_cos * m[0]) + (-trig_sin * m[4]);
    temp1 = (trig_cos * m[1]) + (-trig_sin * m[5]);
    temp2 = (trig_cos * m[2]) + (-trig_sin * m[6]);

    m[4] = (trig_sin * m[0]) + (trig_cos * m[4]);
    m[5] = (trig_sin * m[1]) + (trig_cos * m[5]);
    m[6] = (trig_sin * m[2]) + (trig_cos * m[6]);

    m[0] = temp0;
    m[1] = temp1;
    m[2] = temp2;
}

void f_XformVector(fMatrix m, fVector sv, fVector dv)
{
    float sv0, sv1, sv2;

    sv0 = *sv++;
    sv1 = *sv++;
    sv2 = *sv;

    *dv++ = (m[0] * sv0) + (m[1] * sv1) +
            (m[2] * sv2) + m[3];
    *dv++ = (m[4] * sv0) + (m[5] * sv1) +
            (m[6] * sv2) + m[7];
    *dv   = (m[8] * sv0) + (m[9] * sv1) +
            (m[10] * sv2) + m[11];
}
