// note les dimensions de la texture doivent tre une puissance
// de 2

#include <math.h>

#include "effets.h"
#include "f3Dutils.h"
#include "landscape.h"

// Simplification d'ecriture  mettre ailleurs ventuellement...
Uint32 Landscape::ConvertPixel(const Uint32 src,  const Uint32 srcformat, 
				  const Uint32 srcshift, const Uint32 SHIFT)
{
	const Uint32 ret = ( (src&srcformat) >> srcshift ) << SHIFT;
	return ret;
}

// Charge texture 32bit
void Landscape::LoadTexture(const char *path, SDL_Surface *&tex)
{
  SDL_Surface *textmp = IMG_Load(path);
  if( !textmp ){
    printf("Couldn't open ... %s\n",path ); 
    fflush(stdout);
  }

  assert (textmp->format != 0);
  printf("%d\n", textmp->format->BytesPerPixel);
  assert (textmp->format->BytesPerPixel == 4);

  const int W = textmp->w;
  const int H = textmp->h;

  tex = SDL_CreateRGBSurface (SDL_SWSURFACE, W, H, 32,0xFF0000,
			      0xFF00, 0xFF,0xFF000000);

  LOCK_SURFACE(textmp);
  LOCK_SURFACE(tex);
  Uint32 *src = (Uint32 *)textmp->pixels;
  Uint32 *dst = (Uint32 *)tex->pixels;

  for (int i=0; i<textmp->w*textmp->h; i++) {
    SDL_PixelFormat &sf = *(textmp->format);

    int point = 
      ConvertPixel(src[i], sf.Rmask, sf.Rshift, RSHFT) + 
      ConvertPixel(src[i], sf.Gmask, sf.Gshift, GSHFT) +
      ConvertPixel(src[i], sf.Bmask, sf.Bshift, BSHFT) +
      ConvertPixel(src[i], sf.Amask, sf.Ashift, ASHFT);
    
    dst[i] = point;
    
    assert (dst != 0);
  }
  UNLOCK_SURFACE(textmp);
  UNLOCK_SURFACE(tex);
  SDL_FreeSurface(textmp);
}

// col: 256 couleurs
void Landscape::LoadTexture8(const char *path, 
			     uchar *&tex, Uint32 *col)
{
  SDL_Surface *textmp = IMG_Load(path);
  if ( !textmp ) {
    printf("Couldn't open... %s\n", path);
    fflush(stdout);
  }
  
  const int W = textmp->w;
  const int H = textmp->h;

  tex = new uchar[W*H];
  LOCK_SURFACE(textmp);
  uchar *pix = (uchar *) textmp->pixels;
  for (int i=0; i<W*H; i++) {
    tex[i] = pix[i];
  }
  assert (textmp->format->palette->ncolors <= 256);
  const int FIN = textmp->format->palette->ncolors;
  for (int i=0; i<FIN; i++) {
    SDL_Color c = textmp->format->palette->colors[i];
    col[i] = (c.r<<RSHFT) + (c.g<<GSHFT) + (c.b<<BSHFT);
  }
  for (int i=FIN; i<256; i++) {
    col[i] = 0;
  }
  UNLOCK_SURFACE(textmp);
  SDL_FreeSurface(textmp);
}

void Landscape::prep_uv()
{
  int off2 = 0;
  yangle = 0; //-ftime*T + M_PI;
  x0 = 20;
  y0 = 0; //time*0.01;
  height = 257;
  R = 10*(255*4);

  for (int y=0; y<280; y++) {
    for (int x=0; x<640; x++) {
      fVector v;
      v[0] = x-320;
      v[1] = 290 - y;
      v[2] = 10;
      f_NormalizeVect(v);
      float B = v[0]*x0 + v[1]*height + v[2]*y0;
      float K = x0*x0 + y0*y0 + height*height - R*R;

      float delta2 = B*B - K;
      float delta = sqrt(delta2);

      float t = delta - B;
      f_ScaleVect(v, t);
      float tu1 = atan2(v[0], v[1]);
      float tu2 = (tu1*320.0)/M_PI;
      tu[off2] = tu2;
      off2++;
    }
  }
}

void Landscape::palais(Uint32 *buffer)
{
  LOCK_SURFACE(palais);
  Uint32 *pic = (Uint32 *) spalais->pixels;
  for (int i=0; i<640*480; i++) {
    buffer[i] = pic[i];
  }
  UNLOCK_SURFACE(palais);
}

// Constructeur
Landscape::Landscape()
{
  LoadTexture(PATH_2D "hmap.png", texture);
  LoadTexture8(PATH_2D "texture2.png", tmap, col);
  LoadTexture(PATH_2D "ciel_mozaic.png", skyt);
  LoadTexture(PATH_2D "palais.png", spalais);
  const int S = 256*256;
  hmap = new uchar[S];

  LOCK_SURFACE(texture);
  Uint32 *pix = (Uint32 *) texture->pixels;
  for (int i=0; i<S; i++) {
    hmap[i] = (((pix[i]&BMASK)>>BSHFT)+
      ((pix[i]&GMASK)>>GSHFT) +((pix[i]&RMASK)>>RSHFT) ) / 3;
  }
  UNLOCK_SURFACE(texture);

  ymap = new int[SCREENWIDTH];
  iymap = new int[SCREENWIDTH];
  ixmap = new int[SCREENWIDTH];
  trace = new bool[SCREENWIDTH];
  prep_uv();
}

// Destructeur
Landscape::~Landscape()
{
  SDL_FreeSurface(texture);
  SDL_FreeSurface(skyt);
  delete[] tmap;
  delete[] hmap;
  delete[] ymap;
  delete[] iymap;
  delete[] ixmap;
  delete[] trace;
}

void Landscape::go(Uint32 *screen, Uint32 time)
{
  this->screen = screen;
  //float ftime = (float) time;
  //float T = M_PI / 10000.0;
  
  yangle = 0; //-ftime*T + M_PI;
  x0 = (time/30)&(255*4);
  y0 = 0; //time*0.01;
  height = 257;

  R = 10*(255*4);

  sky(time);
  View();
}


void Landscape::View()
{
  float xvl = cos (yangle - FOV);
  float yvl = sin (yangle - FOV);
  float xvr = cos (yangle + FOV);
  float yvr = sin (yangle + FOV);

  for (int x=0; x<SCREENWIDTH; x++) {
    ymap[x] = SCREENHEIGHT;
    trace[x] = false; // au dbut n'interpole pas
  }

  for (float t=20; t<800; t+=1 +t/64.0) {
    float xpl = x0 + xvl*t;
    float ypl = y0 + yvl*t;
    float xpr = x0 + xvr*t;
    float ypr = y0 + yvr*t;
    int ixpl = (int) (xpl*256.0);
    int iypl = (int) (ypl*256.0);
    int ixpr = (int) (xpr*256.0);
    int iypr = (int) (ypr*256.0);
    line(ixpl, iypl, ixpr, iypr, (int) t);
  }
}

// 24:8
void Landscape::line(int xpl, int ypl, 
		     int xpr, int ypr, int t)
{
  int ixpl2 = xpl; 
  int iypl2 = ypl; 
  int ixpr2 = xpr; 
  int iypr2 = ypr; 

  int invt = 65536/t;
  int iheight = (int) height;

  const int invSW = 1024/SCREENWIDTH;

  for (int x=0; x<SCREENWIDTH; x++) {
    int ix = ixpl2*SCREENWIDTH + x*(ixpr2-ixpl2);
    int iy = iypl2*SCREENWIDTH + x*(iypr2-iypl2);

    ix = (ix * invSW) >> 10;
    iy = (iy * invSW) >> 10;

    const int dix = ix & 0xFF;
    const int diy = iy & 0xFF;
    const int idix = 255 - dix;
    const int idiy = 255 - diy;

    ix = ix >> 8;
    iy = iy >> 8;

    int hix = ix; // ix et iy
    int hiy = iy; // non clipps

    ix &= 0xFF;
    iy &= 0xFF;

    const int py1 = iy<<8;
    const int py2 = ((iy+1)&0xFF)<<8;
    const int px2 = (ix+1)&0xFF;

    const int p1 = hmap[ix+py1];
    const int p2 = hmap[px2+py1];
    const int p3 = hmap[ix+py2];
    const int p4 = hmap[px2+py2];
    const int h = 
      (p1*idix*idiy + p2*dix*idiy + 
       p3*idix*diy + p4*dix*diy)>>16;

    int y = - ( ( (h - iheight) ) * invt ) >> 10; // *64
    y += SCREENHEIGHT/2;

    if (y>=ymap[x]) {
      trace[x] = false;
      continue;
    }

    if (y>=0 && y<480) {
      if (trace[x] == false ) {
	iymap[x] = hiy;
	ixmap[x] = hix;
	trace[x] = true;
      }
      int off2 = x+ymap[x]*SCREENWIDTH;
      assert (ymap[x] != y);
      int diffy = (ymap[x] - y);
      int idiffy = 65536/diffy;
      int t = 0, dt = diffy;

      int hiyt = hiy*t; int iymap_x_dt = iymap[x]*dt;
      int hixt = hix*t; int ixmap_x_dt = ixmap[x]*dt;
      for (int i=ymap[x]; i>y; i--) {
	int iyf = ( (hiyt + iymap_x_dt )*idiffy ) >> 8; // >> 16
	int ixf = ( (hixt + ixmap_x_dt )*idiffy ) >> 16;
	//int cf = (pixels[ ((iyf&0xFF00))+ (ixf&0xFF) ]);
	uchar c = tmap[ ((iyf&0xFF00))+ (ixf&0xFF) ];
	Uint32 cf = col[c];
	screen[off2] = cf;
	off2 -= SCREENWIDTH;
	hiyt += hiy; hixt += hix;
	iymap_x_dt -= iymap[x]; ixmap_x_dt -= ixmap[x];
	t++; dt--;
      }
      iymap[x] = hiy;
      ixmap[x] = hix;
    }
    ymap[x] = y;
  }
}


void Landscape::sky(Uint32 time)
{
  const float T = time*0.01;
  LOCK_SURFACE(skyt);
  Uint32 *psky = (Uint32 *)skyt->pixels;
  int off2 = 0;
  for (int y=0; y<280; y++) {
    for (int x=0; x<640; x++) {
      float tv1 = y;

      float tu2 = tu[off2];
      float tv2 = tv1 + T;

      int iu = ((int) tu2 % 640 ) + 320;
      int iv = ((int) tv2 )&511;

      if (iu>=0 && iu<640 /* && iv>=0 && iv <380 */) {
	screen[off2] = psky[iu + iv*640];
      }
      off2++;
    }
  }
  UNLOCK_SURFACE(skyt);
}
