// plan textur

#include "effets.h"
#include "plan.h"

// Constructeur
Plan::Plan()
{
  // plan textur (chargement de texture)
  SDL_Surface *tmp_text;
  // on charge l'image
  tmp_text = IMG_Load (PATH_2D "terre2.png"); testload (tmp_text);
  // on la fout dans un buffer
  text1 = SDL_CreateRGBSurface (SDL_SWSURFACE, 128, 128, 32,0xFF0000,
				0xFF00, 0xFF,0xFF000000);
  SDL_BlitSurface (tmp_text,NULL, text1, NULL);
  SDL_FreeSurface(tmp_text);

  prep_mipmap();
	
  // camera
  cam0.poscam[0] = 0;       cam0.poscam[1] = 20; cam0.poscam[2] = 0;
  cam0.rotcam[0] = -M_PI/2; cam0.rotcam[1] = 0;  cam0.poscam[2] = 0;
  cam0.focus = 256; cam0.addy = 0;
  // souris
  mouse_rotx = 0; mouse_roty = 0;

  lposcam[0] = lposcam[1] = lposcam[2] = 0;
  lrotcam[0] = lrotcam[1] = lrotcam[2] = 0;
  lposcamadd[0]= lposcamadd[1] = lposcamadd[2] = 0;
  lzrotadd = 0;
}

// Destructeur
Plan::~Plan()
{
  // plan textur
  SDL_FreeSurface(text1); // libre texture

  // detruit map
  for (int i=0; i<nmipmap; i++) {
    delete(mipmap[i]);
  }
}

void Plan::prep_mipmap()
{
  LOCK_SURFACE(text1);
  // mipmap
  Uint32 *src1 = (Uint32 *) text1->pixels;
  int mapsize = 128;
  Uint32 *src = src1;
  for (int i=0; i<nmipmap; i++) {
    mapsize /= 2;
    mipmap[i] = new Uint32[mapsize*mapsize];
    Uint32 *map = mipmap[i];
    for (int y=0; y<mapsize; y++) {
      for (int x=0; x<mapsize; x++) {
	int offset = x*2+y*2*mapsize*2;
	Uint32 blue = ((src[offset]&BMASK) + (src[offset+1]&BMASK) + 
		       (src[offset+mapsize*2]&BMASK) + (src[offset+mapsize*2+1]&BMASK) )/4;
	Uint32 green= ((src[offset]&GMASK) + (src[offset+1]&GMASK) + 
		       (src[offset+mapsize*2]&GMASK) + (src[offset+mapsize*2+1]&GMASK) )/4;
	Uint32 red = ((src[offset]&RMASK) + (src[offset+1]&RMASK) + 
		      (src[offset+mapsize*2]&RMASK) + (src[offset+mapsize*2+1]&RMASK) )/4;
	map[x+y*mapsize] = (red&RMASK) | (green&GMASK) | (blue&BMASK);
      }
    }
    src = map;
  }	
  UNLOCK_SURFACE(text1);
}

// Nvx Mouse qui logge tout comme il faut
void Plan::mouse(SDL_Event event)
{
  switch (event.type)
    {
    case SDL_MOUSEBUTTONDOWN:
      if (event.button.button == 1) {
	ifrot = 1;
      }
      break;
    case SDL_MOUSEBUTTONUP:
      ifrot = 0;
      break;
    case SDL_MOUSEMOTION:
      if (ifrot == 1) {
	if (event.motion.state != 0) {
	  int x_mouse,y_mouse;
	  SDL_GetMouseState(&x_mouse, &y_mouse);
	  lrotcam[1] = (x_mouse*2.0*M_PI)/640.0;
	  lrotcam[0] = -(y_mouse*2.0*M_PI)/480.0;
	}
      }
      break;
    case SDL_KEYDOWN:
      switch (event.key.keysym.sym) {
      case SDLK_UP:
	lposcamadd[2] += 1.0; break;
      case SDLK_DOWN: 
	lposcamadd[2] -= 1.0; break;
      
      case SDLK_SPACE:
	/* cout << "pos " << lposcam[0] << " " ;
	cout << lposcam[1] << " ";
	cout << lposcam[2] << " " << endl;
	cout << "rot " << lrotcam[0] << " " ;
	cout << lrotcam[1] << " ";
	cout << lrotcam[2] << " " << endl; */
	printf("pos %f %f %f\n", 
	       lposcam[0], lposcam[1], lposcam[2]);
	printf("rot %f %f %f\n", 
	       lrotcam[0], lrotcam[1], lrotcam[2]);
	break;
      case SDLK_LEFT:
	lposcamadd[0] -= 1.0; break;
      case SDLK_RIGHT:
	lposcamadd[0] += 1.0; break;
      case SDLK_a: 
	lzrotadd += 0.02; break;
      case SDLK_z:
	lzrotadd -= 0.02; break;
      case SDLK_e:
	lzrotadd = 0; break;
      case SDLK_q: 
	lposcamadd[1] += 1; break;
      case SDLK_s:
	lposcamadd[1] -= 1; break;
      case SDLK_d:
	lposcamadd[1] = 0; break;
      case SDLK_RETURN:
	lposcamadd[0] = lposcamadd[1] = lposcamadd[2] = 0;
	break;
      default:
	break;
      }
      break;
      //    case SDL_KEYUP:
      //if ((event.key.keysym.sym==SDLK_UP)||(event.key.keysym.sym==SDLK_DOWN) ){
      //cam0.addy=0;
      //}
      //break;
    }
}


void Plan::getMouse(fVector lposcam_, fVector lrotcam_)
{
  for (int i=0; i<3; i++) {
    lposcam[i] += lposcamadd[i];
  }
  lrotcam[2] += lzrotadd;
  for (int i=0; i<3; i++) {
    lposcam_[i] = lposcam[i];
    lrotcam_[i] = lrotcam[i];
  }
}

// Calcule les coordonnes de la camera
void Plan::coord(Uint32 time)
{
  float time1 = time*0.001; // vitesse de dfilement du temps

  // position de la camera
  cam0.poscam[0] =  lposcam[0];
  cam0.poscam[1] =  /* lposcam[1]*/ 100;
  cam0.poscam[2] =  lposcam[2]; //100 + 200 * sin(time1); 
  // orientation de la camera
  cam0.rotcam[0] = lrotcam[0];
  cam0.rotcam[1] = lrotcam[1];		
  cam0.rotcam[2] = lrotcam[2];

  /*	if (time<6000) {
	poscamadd[0] = 0;
	poscamadd[1] = CLAMP(time1,0,20);
	poscamadd[2] = 0;
	
	rotcamadd[0] = 0;
	rotcamadd[1] = 0.01*sin(time1);
	rotcamadd[2] = 0;
	}
	else if (time<9540) {
	poscamadd[0] = 0;

	poscamadd[2] = 0;

	rotcamadd[0] = 0.1*sin(time1);
	rotcamadd[1] = CLAMP( (time-6000), 0, 0.1)*0.001;
	rotcamadd[2] = 0.01*sin(time1);
	}
	else if (time<20000) {
	rotcamadd[0] = 0.0001;
	rotcamadd[1] = 0;
	rotcamadd[2] = 0.01*sin(time1);

	poscamadd[1] = 20 - CLAMP( (time-9540)*0.001 ,0,10);
	poscamadd[0] = 20 * sin( (time-9540)*0.0001 );
	poscamadd[2] = -30 * sin( (time-9540)*0.0001 );		
	}
	else if (time<40000) {
	rotcamadd[0] = 0;
	rotcamadd[1] = 0;
	rotcamadd[2] = 0.01*sin(time1);

	poscamadd[1] = 0;
	poscamadd[0] = 0;
	poscamadd[2] = 0;
	}
	else {
	rotcamadd[0] = 0.0002;
	rotcamadd[1] = 0;
	rotcamadd[2] = 0.01*sin(time1);

	poscamadd[1] = 10 - CLAMP( (time-9540)*0.001 ,0,10);
	poscamadd[0] = 200 * sin( (time-9540)*0.0001 );
	poscamadd[2] = -300 * sin( (time-9540)*0.0001 );
	}

	cam0.poscam[0] += poscamadd[0];
	cam0.poscam[1] += poscamadd[1];
	cam0.poscam[2] += poscamadd[2]; 	

	cam0.rotcam[0] += rotcamadd[0];
	cam0.rotcam[1] += rotcamadd[1];
	cam0.rotcam[2] += rotcamadd[2];
  */

}

// On appelle cette fonction pour lancer l'effet
void Plan::plan_txt(Uint32 *buffer)
{
  LOCK_SURFACE (text1);
  Uint32 *src = (Uint32 *) text1->pixels;
  do_map (buffer,src, cam0);
  UNLOCK_SURFACE (text1);
}

void Plan::fillvect (double *vect, double x,double y,double z)
{
  *vect = x;
  *(vect+1) = y;
  *(vect+2) = z;
}

void Plan::normvect (double *vect, double *norm)
{
  double tmp_norm;
  tmp_norm = vect[0]*vect[0] + vect[1]*vect[1] + vect[2]*vect[2]; 
  *norm = sqrt(tmp_norm);
  tmp_norm = 1.0/(*norm);
  vect[0] *= tmp_norm; vect[1] *= tmp_norm; vect[2] *= tmp_norm;
}

void Plan::rot (double *rot, double *vect)
{
  // TODO cos opt: remaque on s'en fou on s'en sert plus
  double rx,ry,rz;

  ry = cos(rot[0])*vect[1] - sin(rot[0])*vect[2];
  rz = sin(rot[0])*vect[1] + cos(rot[0])*vect[2];

  rx = cos(rot[1])*vect[0] + sin(rot[1])*rz;
  vect[2] = cos(rot[1])*rz - sin(rot[1])*vect[0];

  vect[0] = cos(rot[2])*rx - sin(rot[2])*ry;
  vect[1] = sin(rot[2])*rx  + cos(rot[2])*ry;

}

void Plan::rotate_cam (double *dircamx, double *dircamy, 
		       double *dircamz, T_camera &cam)
{

  /*	f_make_identity_matrix( rm );
	f_append_zrot_matrix( rm, cam.rotcam[2]);
	f_append_yrot_matrix( rm, cam.rotcam[1]); // ****** OK *******
	f_append_xrot_matrix( rm, cam.rotcam[0]);*/
	
  f_make_xyzrot_matrix (rm, cam.rotcam[0], cam.rotcam[1], cam.rotcam[2]);

  dircamx[0] = rm[0]; dircamy[0] = rm[1]; dircamz[0] = rm[2];
  dircamx[1] = rm[4]; dircamy[1] = rm[5]; dircamz[1] = rm[6];
  dircamx[2] = rm[8]; dircamy[2] = rm[9]; dircamz[2] = rm[10];

}

void Plan::putpixel1 (Uint32 *src, Uint32 *dest, int offsetd2,
		      Uint32 xf, Uint32 yf, Uint32 lf)
{
  Uint32 pixel;
  Uint32 r,g,b;
  pixel = src[xf+yf];
  r = pixel & RMASK;
  g = pixel & GMASK;
  b = pixel & BMASK;
  r = (r * lf ) >> 8;
  g = (g * lf ) >> 8;
  b = (b * lf ) >> 8;
  pixel = (r & RMASK) + (g & GMASK) + (b & BMASK);
  dest[offsetd2] = pixel;
}

void Plan::putpixel2 (Uint32 *src, Uint32 *dest, int offsetd2,
		      Uint32 xf, Uint32 yf, Uint32 lf)
{
  Uint32 dx,dy,cdx,cdy;
  Uint32 pix;
  Uint32 r1,g1,b1,r2,g2,b2,r3,g3,b3,r4,g4,b4;
	
  dx = xf & 255; dy = yf & 255;
  cdx = 255 - dx; cdy = 255 - dy;
  xf = xf >> 8; yf = yf >> 8;
  xf &= 127; yf &= 127;
  yf = yf << 7;
  pix = src[xf+yf];
  r1 = (pix & RMASK)>>16; g1 = (pix & GMASK)>>8; b1 = pix & BMASK;
  pix = src[((xf+1)&127)+yf];
  r2 = (pix & RMASK)>>16; g2 = (pix & GMASK)>>8; b2 = pix & BMASK;
  pix = src[xf+((yf+128)&(127<<7))];
  r3 = (pix & RMASK)>>16; g3 = (pix & GMASK)>>8; b3 = pix & BMASK;
  pix = src[((xf+1)&127) + ((yf+128)&(127<<7))];
  r4 = (pix & RMASK)>>16; g4 = (pix & GMASK)>>8; b4 = pix & BMASK;		
  r1 = ((r1*cdx*cdy + r2*dx*cdy + r3*dy*cdx + r4*dx*dy)>>16);
  g1 = ((g1*cdx*cdy + g2*dx*cdy + g3*dy*cdx + g4*dx*dy)>>16);	
  b1 = ((b1*cdx*cdy + b2*dx*cdy + b3*dy*cdx + b4*dx*dy)>>16);
  r1 = r1<<16; g1 = g1<<8;
  r1 = (r1 * lf ) >> (8+0);
  g1 = (g1 * lf ) >> (8+0);
  b1 = (b1 * lf ) >> (8+0);
  pix = (r1 & RMASK) | (g1 & GMASK) | (b1 & BMASK);
  dest[offsetd2] = pix;
}

void Plan::interp(int x1, int x2, int x3, int x4, 
		  int y1, int y2, int y3, int y4,
		  int l1, int l2, int l3, int l4, 
		  int offsetd1,
		  Uint32 *dest, Uint32 *src)
{
  int x,y, xf, yf, lf, xg, xd, yg,yd, lg, ld;
  int dx,dy;
  int offsetd2 = offsetd1;
	
  if ((l1 == 0) && (l2 == 0) && (l3 == 0) && (l4 == 0) ) {
    for (y=0; y<STEP; y++) {
      for (x=0; x<STEP; x++) { dest[offsetd2++] = 0; }
      offsetd2 -= STEP; offsetd2 += 640;
    }
    return;
  }
  for (y=0, dy = 0; y<STEP; y++, dy++) {
    xg = (x1 * (STEP-dy) + x3 * dy) >> (STEP_SHIFT+SHIFT1);
    xd = (x2 * (STEP-dy) + x4 * dy) >> (STEP_SHIFT+SHIFT1);
    yg = (y1 * (STEP-dy) + y3 * dy) >> (STEP_SHIFT+SHIFT1);
    yd = (y2 * (STEP-dy) + y4 * dy) >> (STEP_SHIFT+SHIFT1);
    lg = (l1 * (STEP-dy) + l3 * dy) >> STEP_SHIFT;
    ld = (l2 * (STEP-dy) + l4 * dy) >> STEP_SHIFT;
    for (x=0, dx=0; x<STEP; x++, dx++) {
      xf = ( xg * (STEP-dx) + xd * dx );
      yf = ( yg * (STEP-dx) + yd * dx );
      lf = ( lg * (STEP-dx) + ld * dx) >> STEP_SHIFT;
      if (lf>=180) { 
	// bilinear filtering
	xf = xf >> ((STEP_SHIFT+SHIFT2)-8);
	yf = yf >> ((STEP_SHIFT+SHIFT2)-8);
	putpixel2 (src, dest, offsetd2, xf, yf, lf);
      }
      else {
	int wmap = (180-lf)>>6;
	if (wmap>nmipmap-1) wmap=nmipmap-1;
	if (wmap<0) wmap=0;
	// mipmapping
	xf = xf >> (STEP_SHIFT+SHIFT2+1+wmap);
	yf = yf >> (STEP_SHIFT+SHIFT2+1+wmap);
	xf &= (1<<(6-wmap))-1;
	yf &= (1<<(6-wmap))-1;
	yf = yf << (6-wmap);
	putpixel1 (mipmap[wmap], dest, offsetd2, xf, yf, lf);
      }
      offsetd2++;
    }
    offsetd2 -= STEP;
    offsetd2 += 640; 
  }
}

void Plan::do_map(Uint32 *dest, Uint32 *src, T_camera &cam)
{
  int x,y;
  float xs,ys;
  int offsetd = 0;
  double dircamx[3], dircamy[3], dircamz[3]; // pour le lancer de rayon
  double ray[3],ray2[3];	// un rayon
  double norm;		// stoque la longueur du rayon	
  double t;
  int i, x2, y2;
  float tmp;

  fillvect(dircamx,1,0,0); 
  fillvect(dircamy,0,1,0);
  fillvect(dircamz,0,0,cam.focus);  // * focal
  rotate_cam (dircamx, dircamy, dircamz, cam);
  for(int i=0; i<3;i++) dircamz[i] *= cam.focus;
	
  ray[0] = dircamz[0] - dircamy[0]*240 - dircamx[0]*320;
  ray[1] = dircamz[1] - dircamy[1]*240 - dircamx[1]*320;
  ray[2] = dircamz[2] - dircamy[2]*240 - dircamx[2]*320;
  int uvoff=0;  
  for (y2=0; y2<HAUTEUR; y2++) {
    for (x2=0; x2<LARGEUR; x2++) {
      for (i=0; i<3; i++) { ray2[i] = ray[i];}
      tmp = 0;
      if (fabs(ray2[1]) != 0) {
	normvect (ray2, &norm);		// on normalize
	t = cam.poscam[1]/ray2[1];	
	if (t>0) {
	  xs = ( 0.05*(cam.poscam[0] + t*ray2[0])	);
	  ys = ( -0.06*(cam.poscam[2] + t*ray2[2]) );
	  t /= ray2[1]*ray2[1]*256.0;
	  if (t>255) t=255;
	  t = 255-t;
	  uvprecalc[uvoff] = xs;
	  uvprecalc[uvoff+1] = ys;
	  tmp = t;
	} 
      }
      uvprecalc[uvoff+2] = tmp; 
      ray[0] += dircamx[0]*STEP;
      ray[1] += dircamx[1]*STEP;
      ray[2] += dircamx[2]*STEP;
      uvoff += 3;
    }
    ray[0] +=  dircamy[0]*STEP - dircamx[0]*(640+STEP) ;
    ray[1] +=  dircamy[1]*STEP - dircamx[1]*(640+STEP) ;
    ray[2] +=  dircamy[2]*STEP - dircamx[2]*(640+STEP) ;
  }
  // interpolation
  for (y=0, y2=0; y<480; y+=STEP, offsetd += STEP*640, y2 += LARGEUR) {
    for (x=0, x2=0; x<640; x+=STEP, x2++) {
      Uint32 calc = (x2+y2)*3;
      float pt1,pt2,pt3,pt4;
      float px1,px2,px3,px4;
      float py1,py2,py3,py4;
      pt1 = uvprecalc[calc+2]; px1 = uvprecalc[calc]*MUL1; py1 = uvprecalc[calc+1]*MUL1;
      calc += 3;
      pt2 = uvprecalc[calc+2]; px2 = uvprecalc[calc]*MUL1; py2 = uvprecalc[calc+1]*MUL1;
      calc += LARGEUR*3;
      pt4 = uvprecalc[calc+2]; px4 = uvprecalc[calc]*MUL1; py4 = uvprecalc[calc+1]*MUL1;
      calc -= 3;
      pt3 = uvprecalc[calc+2]; px3 = uvprecalc[calc]*MUL1; py3 = uvprecalc[calc+1]*MUL1;
      interp((int)px1, (int)px2, (int) px3, (int) px4, 
	     (int) py1, (int) py2, (int) py3, (int) py4,
	     (int) pt1, (int) pt2, (int) pt3, (int) pt4, 
	     offsetd+x, dest, src); 
    }
  }
}


void Plan::do_lignes( Uint32 *screen )
{
  int i;
  fMatrix m[2];
  fMatrix tm[2];
  fVector ntri[2];

  f_transpose_matrix(rm);

  fVector myline[2][2] = { { {     0.0,   0.0,   500.0 },
			     {  5000.0,   0.0,   500.0 } },
			   { {   500.0,   0.0,     0.0 },
			     {   500.0,   0.0,  5000.0 } } };

  for( double bclz = 0.0 ; bclz <= 50.0 ; bclz += 10.0 ) {
    for( double bcl = -500.0 ; bcl <= 5000.0 ; bcl += 100.0 ) {
	
      f_make_identity_matrix( tm[0] );
      f_make_identity_matrix( tm[1] );
	
      f_trans_matrix( tm[0], -cam0.poscam[0]/10.0, cam0.poscam[1]/10.0-10.0-bclz, 
		      -cam0.poscam[2]/10.0 + bcl -500.0 );
      f_trans_matrix( tm[1], -cam0.poscam[0]/10.0 + bcl, cam0.poscam[1]/10.0-10.0-bclz, 
		      -cam0.poscam[2]/10.0 -500.0);

      f_ConcatMatrices( rm, tm[0], m[0]);
      f_ConcatMatrices( rm, tm[1], m[1]);
			
      for( int num=0; num<2; num++ ) {
	for( i=0; i<2; i++ ){
	  f_XformVector( m[num], myline[num][i], ntri[i] );
	}
	
	double Ax = ntri[0][VECT_X];
	double Ay = ntri[0][VECT_Y];
	double Az = ntri[0][VECT_Z];
	
	double Bx = ntri[1][VECT_X];
	double By = ntri[1][VECT_Y];
	double Bz = ntri[1][VECT_Z];

	// CLIP EN Z

	if( Az < -250.0 && Bz < -250.0 ) continue;
	else if ( Az < -250.0 ) {
	  if(Bz-Az < 1.0) continue;
	  double coupeZ = -(Az + 250.0);
	  double ratioZ = coupeZ/(Bz - Az);
	  Az += (Bz-Az) * ratioZ;
	  Ax += (Bx-Ax) * ratioZ;
	  Ay += (By-Ay) * ratioZ;
	} else if( Bz < -250.0 ) {
	  if(Az-Bz < 1.0) continue;
	  double coupeZ = -(Bz + 250.0);
	  double ratioZ = coupeZ/(Az - Bz);
	  Bz += (Az-Bz) * ratioZ;
	  Bx += (Ax-Bx) * ratioZ;
	  By += (Ay-By) * ratioZ;
	}

	Ax *= 255.0/(Az+255.0);
	Ay *= 255.0/(Az+255.0);
	Bx *= 255.0/(Bz+255.0);
	By *= 255.0/(Bz+255.0);

	// CLIP en X gauche
	if( Ax < -320.0 ) {
	  if( Bx <= Ax ) continue;
	  double coupeX = -(Ax + 320);
	  double ratioX = coupeX/(Bx-Ax);
	  Ax += (Bx-Ax) * ratioX;
	  Ay += (By-Ay) * ratioX;
	}
	if( Ax > 320.0 ) {
	  if( Bx >= Ax ) continue;
	  double coupeX = (Ax - 320);
	  double ratioX = coupeX/(Ax-Bx);
	  Ax += (Bx-Ax) * ratioX;
	  Ay += (By-Ay) * ratioX;
	}
	if( Bx < -320.0 ) {
	  if( Ax <= Bx ) continue;
	  double coupeX = -(Bx + 320);
	  double ratioX = coupeX/(Ax-Bx);
	  Bx += (Ax-Bx) * ratioX;
	  By += (Ay-By) * ratioX;
	}
	if( Bx > 320.0 ) {
	  if( Ax >= Bx ) continue;
	  double coupeX = (Bx - 320);
	  double ratioX = coupeX/(Bx-Ax);
	  Bx += (Ax-Bx) * ratioX;
	  By += (Ay-By) * ratioX;
	}

	if( Ay < -240.0 ) {
	  if( By <= Ay ) continue;
	  double coupeY = -(Ay + 240);
	  double ratioY = coupeY/(By-Ay);
	  Ay += (By-Ay) * ratioY;
	  Ax += (Bx-Ax) * ratioY;
	}
	if( Ay > 240.0 ) {
	  if( By >= Ay ) continue;
	  double coupeY = (Ay - 240);
	  double ratioY = coupeY/(Ay-By);
	  Ax += (Bx-Ax) * ratioY;
	  Ay += (By-Ay) * ratioY;
	}
	if( By < -240.0 ) {
	  if( Ay <= By ) continue;
	  double coupeY = -(By + 320);
	  double ratioY = coupeY/(Ay-By);
	  Bx += (Ax-Bx) * ratioY;
	  By += (Ay-By) * ratioY;
	}
	if( By > 240.0 ) {
	  if( Ay >= By ) continue;
	  double coupeY = (By - 320);
	  double ratioY = coupeY/(By-Ay);
	  Bx += (Ax-Bx) * ratioY;
	  By += (Ay-By) * ratioY;
	}

	int couleur = ((int)bclz*5)& 0xFF; 
	line2(screen, (int)Ax+320, (int)Ay+240,
	      (int)Bx+320, (int)By+240,
	      0xFF000000 | (couleur<<8) | ((255-couleur)<<16) );
      }

    }
  }
}
