
%{
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "timeparse.h"

extern "C"
{
	int yyparse(void);
	int yylex(void); 

	void yyerror(const char* str)
	{
		fprintf(stderr, "error: %s\n",str);
	}

	int yywrap()
	{
		return 1;
	}
}

extern FILE* yyin;
Timeparse *me;

Timeparse::Timeparse(char *file)
{
  me = this;
  yyin = fopen (file, "r");
  if (yyin == 0) {
    fprintf(stderr, "erreur chargement %s\n", file);
  }
  yyparse();
}

%}

%token WORD NUMBER POS FLOAT

%%
commands: /* empty */
	| commands command
	;

command:
WORD NUMBER NUMBER
{
  char *word = (char*) $1;
  int debut = $2;
  int duree = $3;
#ifdef DEBUG_TPARSE
  printf ("function: %s, ", word);
  printf ("debut: %d, dure: %d\n", debut, duree);
#endif
  string function = word;
  
  Func func(function, debut, duree);
  me->lfunc.push_back(func);
}
| position
{ }
;

position: POS FLOAT FLOAT FLOAT
{
  char *f = (char *)$1;
  float a = (float) $2;
  float b = (float) $3;
  float c = (float) $4;
#ifdef DEBUG_TPARSE
  printf ("%s %f %f %f\n", f, a, b, c);
#endif
  const string name = f;
  const string spos = "pos";
  const string srot = "rot";

  Vect vect;
  vect.v[0] = a; vect.v[1] = b; vect.v[2] = c;

  if (f == spos) {
    me->vpos.push_back(vect);
  }
  else if (f == srot) {
    me->vrot.push_back(vect);
  }
}
;


%%

bool Timeparse::findTime(const int time, 
	     list<Func>::iterator &ifunc )
{
  while ( ifunc != lfunc.end() ) {
    TTime &timing = ifunc->timing;
    if (time >= timing.debut && time < timing.debut+timing.duree) {
      return true;
    }
    ++ifunc;
  }
  return false;
}
