// test de perlin noise style

#include "effets.h"

void gen_pal(Uint32 *pal);

// constructeur
Z1::Z1()
{
	SIZ2 = 7;
	SIZE = (1<<SIZ2);
	FOCUS = 256;
	surf1 = SDL_CreateRGBSurface (SDL_SWSURFACE, SIZE, SIZE, 8, 0, 0, 0, 0);
	surf2 = SDL_CreateRGBSurface (SDL_SWSURFACE, SIZE, SIZE, 8, 0, 0, 0, 0);
	surf3 = SDL_CreateRGBSurface (SDL_SWSURFACE, SIZE, SIZE, 8, 0, 0, 0, 0);
	surf4 = SDL_CreateRGBSurface (SDL_SWSURFACE, SIZE, SIZE, 8, 0, 0, 0, 0);
	fin1 = SDL_CreateRGBSurface (SDL_SWSURFACE, SIZE, SIZE, 8, 0, 0, 0, 0);

	CX = (640 - SIZE)/2;
	CY = (480 - SIZE)/2;

	LOCK_SURFACE(surf1);
	LOCK_SURFACE(surf2);
	LOCK_SURFACE(surf3);
	LOCK_SURFACE(surf4);
	LOCK_SURFACE(fin1);
	uchar *src1 = (uchar *) surf1->pixels;
	uchar *src2 = (uchar *) surf2->pixels;
	uchar *src3 = (uchar *) surf3->pixels;
	uchar *src4 = (uchar *) surf4->pixels;
	uchar *pfin = (uchar *) fin1->pixels; 
	perlin(src1,32,128);
	perlin(src2,16,64);	
	perlin(src3,8,32);
	perlin(src4,4,16);

	// gnre la text fin
	somme(pfin, src1, src2, src3, src4);
	
	UNLOCK_SURFACE(fin1);
	UNLOCK_SURFACE(surf1);
	UNLOCK_SURFACE(surf2);
	UNLOCK_SURFACE(surf3);
	UNLOCK_SURFACE(surf4);

	SDL_FreeSurface(surf1);
	SDL_FreeSurface(surf2);
	SDL_FreeSurface(surf3);
	SDL_FreeSurface(surf4);

	// racines
	racine();

  tz=256;
	step = 4;


	// genere la palette
	gen_pal(pal1);

}

void gen_pal(Uint32 *pal)
{
	int off = 0;
	for (int i=0; i<255; i++) {
		pal[off++] = i;
	}
	for (int i=0; i<255; i++) {
		pal[off++] = (255-i/4) + (CLAMP(i,0,128)<<16);
	}
	for (int i=0; i<255; i++) {
		pal[off++] = CLAMP(192+i/2,0,255) + ((i/2)<<8) + ((128+i/4)<<16)  ;
	}	
	for (int i=0; i<255; i++) {
		pal[off++] = 255-i + ((128-i/2)<<8) + (CLAMP(192-i,0,255)<<16)  ;
	}	
}

// destructeur
Z1::~Z1(void)
{
	SDL_FreeSurface(fin1);
}

void Z1::go(Uint32 *buffer, Uint32 time)
{
	//	Uint32 off1 = CX + CY*640;
	const int DEPTH_MAX = 256;
	const int DEPTH_STEP = 128;
	const float ftime = time;
	const int time2 = (int) ( 0.04*ftime  );
	float ftime3 = 0.0001*ftime;
	int dec1 = (int) ( 100.0 * sin(ftime3)     );
	int dec2 = (int) ( 128.0 * cos(ftime3+0.7) );	

	LOCK_SURFACE(fin1);
	uchar *pfin = (uchar *) fin1->pixels;

	tz+=step; 
	if (tz>256) {step = -step;}
	if (tz<32) {step = -step; fflush(stdout);}

	Uint32 off = 200*640+32;
	for (int y=200; y<480-200; y++) {
		for (int x=32; x<640-32; x++) {

			Uint32 color = 0;

			int tz2 = (int) (tz*512.0);
			int depthmz = (DEPTH_STEP * pcalcz[off])>>9;
			int depth = (tz2 * pcalcz[off])>>9;

			int tx=(pcalcx[off]*depth)>>9;
			int ty=(pcalcy[off]*depth)>>9;

			int idepthmx = (pcalcx[off]*depthmz);
			int idepthmy = (pcalcy[off]*depthmz);
			
			for (int d2=DEPTH_MAX; d2>0; d2 -= DEPTH_STEP) { 
				tx -= idepthmx;
				ty -= idepthmy;

				int u = (int) (tx>>9) + dec1;
				int v = (int) (ty>>9) + dec2;
				u = u & 127;
				v = v & 127;
				int ctmp = pfin[u+(v<<7)];
				
				color += ctmp - (255-d2);
			} 
			int col3 = (int) time2 + color;
			buffer[off] = pal1 [col3&1023];

			off++;		
		}
		off+=64;
	}

	UNLOCK_SURFACE(fin1);

	asm_emms();
}

void Z1::perlin(uchar *src, const int step=32, const int RANGE=256)
{
	for (int y=0; y<=SIZE-step; y+=step) {
		for (int x=0; x<=SIZE-step; x+=step) {
			src[x+(y<<SIZ2)] = rand()%(RANGE);
		}
	}

	// interpole
//	Uint32 off=0;
	for (int y=0; y<=SIZE-step; y+=step) {
		for (int x=0; x<=SIZE-step; x+=step) {
			Uint32 xm = x%(SIZE), ym = y%(SIZE);
			Uint32 xm2 = (x+step)%(SIZE), ym2 = (y+step)%(SIZE);
			uchar C1 = src[xm+(ym<<SIZ2)];
			uchar C2 = src[xm2+(ym<<SIZ2)];
			uchar C3 = src[xm+(ym2<<SIZ2)];
			uchar C4 = src[xm2+(ym2<<SIZ2)];
			// interp
			for (int y1=0; y1<step; y1++) {
				int iy1=step-y1;
				for (int x1=0; x1<step; x1++) {
					int ix1=step-x1;
					Uint32 C = (C1*ix1*iy1 + C2*x1*iy1
									  +C3*ix1*y1  + C4*x1*y1)/(step*step);
					src[x+x1+((y+y1)<<SIZ2)] = (uchar) C;
				}
			}
		}
	}
}

void Z1::somme(uchar *buffer, uchar *src1, uchar *src2, uchar *src3, 
								uchar *src4)
{
	int off = 0;
	for (int y1=0; y1<SIZE; y1++) {
		for (int x1=0; x1<SIZE; x1++) {
			Uint32 a = src1[off] + src2[off] +
								 src3[off] + src4[off];
			a = CLAMP(a,0,255);
			buffer[off++] = a;
		}
	}
}

// normalise les vect
void Z1::racine()
{
	int off=0;
	for (int y=0; y<480; y++) {
		for (int x=0; x<640; x++) {
			double ix = (x-320), iy = (y-240), iz = FOCUS; 
			double lenght = ix*ix + iy*iy + iz*iz;
			double invl = sqrt(lenght);
			pcalcx[off] = (int) ( 512.0 * ix / invl ); 
			pcalcy[off] = (int) ( 512.0 * iy / invl ); 
			pcalcz[off] = (int) ( 512.0 * invl / iz ); 
			off++;
		}
	}
}

// end of Z1.cc
